# base_stream.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Thu May 15 16:57:27 UTC 2025

"""Base class for all stream controllers."""

from __future__ import annotations

import enum
import logging
from abc import ABC, abstractmethod
from typing import Generic, TypeVar

from returns.result import ResultE

from open_gopro.gopro_base import GoProBase
from open_gopro.models.streaming import StreamType

logger = logging.getLogger(__name__)

T = TypeVar("T")


class StreamController(ABC, Generic[T]):
    """Interface for stream controllers."""

    class StreamStatus(enum.Enum):
        """Enum for the different types of stream status."""

        STARTING = enum.auto()
        STARTED = enum.auto()
        STOPPING = enum.auto()
        STOPPED = enum.auto()

    def __init__(self, gopro: GoProBase) -> None:
        self.gopro = gopro

    @property
    @abstractmethod
    def is_available(self) -> bool:
        """Check if the stream is available.

        Returns:
            bool: True if the stream is available, False otherwise.
        """

    @property
    @abstractmethod
    def stream_type(self) -> StreamType:
        """Get the type of the stream.

        Returns:
            StreamType: The type of the stream.
        """

    @property
    @abstractmethod
    def url(self) -> str:
        """Get the URL that can be used to view the stream

        Returns:
            str: The URL of the stream.
        """

    @property
    @abstractmethod
    def status(self) -> StreamStatus:
        """Get the current status of the stream.

        Returns:
            StreamStatus: The current status of the stream.
        """

    @abstractmethod
    async def start(self, options: T) -> ResultE[None]:
        """Start the stream.

        Args:
            options (T): The options to use for the stream.

        Returns:
            ResultE[None]: Result container with failure if it exists.
        """

    @abstractmethod
    async def stop(self) -> ResultE[None]:
        """Stop the stream."""

    @abstractmethod
    async def close(self) -> ResultE[None]:
        """Close the stream controller, gracefully cancelling tasks and releasing resources.

        Returns:
            ResultE[None]: Result container with failure if it exists.
        """



# streaming.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Thu May 15 16:57:28 UTC 2025

"""Streaming models and entities."""

import enum
from dataclasses import dataclass
from pathlib import Path

from pydantic import Field

from open_gopro.domain.enum import GoProEnum, GoProIntEnum
from open_gopro.models.bases import CustomBaseModel
from open_gopro.models.general import SupportedOption
from open_gopro.models.proto import EnumLens, EnumWindowSize


class StreamType(enum.Enum):
    """Enum for the different types of streams."""

    WEBCAM = enum.auto()
    LIVE = enum.auto()
    PREVIEW = enum.auto()


class WebcamResolution(GoProIntEnum):
    """Possible Webcam Resolutions"""

    NOT_APPLICABLE = 0
    RES_480 = 4
    RES_720 = 7
    RES_1080 = 12


class WebcamFOV(GoProIntEnum):
    """Possible Webcam FOVs"""

    WIDE = 0
    NARROW = 2
    SUPERVIEW = 3
    LINEAR = 4


class WebcamProtocol(GoProEnum):
    """Possible Webcam Protocols"""

    TS = "TS"
    RTSP = "RTSP"


class WebcamStatus(GoProIntEnum):
    """Webcam Statuses / states"""

    OFF = 0
    IDLE = 1
    HIGH_POWER_PREVIEW = 2
    LOW_POWER_PREVIEW = 3


class WebcamError(GoProIntEnum):
    """Errors common among Webcam commands"""

    SUCCESS = 0
    SET_PRESET = 1
    SET_WINDOW_SIZE = 2
    EXEC_STREAM = 3
    SHUTTER = 4
    COM_TIMEOUT = 5
    INVALID_PARAM = 6
    UNAVAILABLE = 7
    EXIT = 8


class WebcamResponse(CustomBaseModel):
    """Common Response from Webcam Commands"""

    status: WebcamStatus | None = Field(default=None)
    error: WebcamError
    setting_id: str | None = Field(default=None)
    supported_options: list[SupportedOption] | None = Field(default=None)


@dataclass
class LivestreamOptions:
    """Livestream-specific stream setup options"""

    url: str
    minimum_bitrate: int = 1000
    maximum_bitrate: int = 1000
    starting_bitrate: int = 1000
    encode: bool = True
    resolution: EnumWindowSize.ValueType | None = None
    fov: EnumLens.ValueType | None = None
    certs: list[Path] | None = None


@dataclass
class PreviewStreamOptions:
    """Preview stream-specific stream setup options"""

    port: int | None = None


@dataclass
class WebcamStreamOptions:
    """Webcam stream-specific stream setup options"""

    resolution: WebcamResolution | None = None
    protocol: WebcamProtocol | None = None
    fov: WebcamFOV | None = None
    port: int | None = None



# preview_stream.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Mon Jul 31 17:04:07 UTC 2023

"""Example to start and view a preview stream"""

import argparse
import asyncio
import logging
import sys

from returns.pipeline import is_successful
from rich.console import Console

from open_gopro import WirelessGoPro
from open_gopro.demos.gui.video_display import BufferlessVideoCapture
from open_gopro.models import streaming
from open_gopro.util import add_cli_args_and_parse
from open_gopro.util.logger import set_stream_logging_level, setup_logging
from open_gopro.util.util import ainput

console = Console()


async def main(args: argparse.Namespace) -> int:
    setup_logging(__name__, args.log)

    async with WirelessGoPro(args.identifier) as gopro:
        set_stream_logging_level(logging.WARNING)

        with console.status("Starting preview stream..."):
            if not is_successful(
                result := (
                    await gopro.streaming.start_stream(
                        stream_type=streaming.StreamType.PREVIEW,
                        options=streaming.PreviewStreamOptions(port=args.port),
                    )
                )
            ):
                console.print(f"[red]Failed to start preview stream: {result.failure()}")
                return 1
            assert gopro.streaming.url, "Preview stream URL should not be empty after starting the stream"

        console.print("Preview stream started. It can be viewed in VLC at [yellow]udp://@:8554")
        console.print(
            "Press Enter to view the preview stream using Python CV2. Once started, it won't be viewable in VLC."
        )
        await ainput("")

        with console.status("Displaying the preview stream..."):
            BufferlessVideoCapture(
                source=gopro.streaming.url,
                protocol=BufferlessVideoCapture.Protocol.TS,
                printer=console.print,
            ).display_blocking()

        with console.status("Stopping preview stream..."):
            await gopro.streaming.stop_active_stream()
        return 0


def parse_arguments() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description="Connect to the GoPro via BLE and Wifi, start a preview stream, then display it with CV2."
    )
    parser.add_argument(
        "--port", type=int, help="Port to use for livestream. Defaults to 8554 if not set", default=8554
    )
    return add_cli_args_and_parse(parser, wifi=False)


def entrypoint() -> None:
    sys.exit(asyncio.run(main(parse_arguments())))


if __name__ == "__main__":
    entrypoint()



# stream_feature.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Thu May 15 16:57:27 UTC 2025

"""Combined stream functionality for all type of video streams."""

from __future__ import annotations

import logging

from returns.result import Failure, ResultE, Success

from open_gopro.domain.exceptions import GoProError
from open_gopro.features.base_feature import BaseFeature
from open_gopro.features.streaming.base_stream import StreamController, StreamType
from open_gopro.features.streaming.livestream import (
    LiveStreamController,
    LivestreamOptions,
)
from open_gopro.features.streaming.preview_stream import (
    PreviewStreamController,
    PreviewStreamOptions,
)
from open_gopro.features.streaming.webcam_stream import (
    WebcamStreamController,
    WebcamStreamOptions,
)
from open_gopro.models.types import StreamOptions

logger = logging.getLogger(__name__)


class StreamFeature(BaseFeature):
    """Abstracted stream functionality for all type of video streams.

    There can only ever be one stream open at a time.
    """

    def __init__(self) -> None:
        super().__init__()
        self._current: StreamController | None = None

    @property
    def is_supported(self) -> bool:  # noqa: D102
        # Streaming is always supported
        return True

    @property
    def current_stream(self) -> StreamType | None:
        """Get the current stream type.

        Returns:
            StreamType | None: stream type if a stream is active, None otherwise.
        """
        return self._current.stream_type if self._current else None

    @property
    def is_streaming(self) -> bool:
        """Check if a stream is currently active.

        Returns:
            bool: True if a stream is active, False otherwise.
        """
        return self._current is not None and self._current.status == StreamController.StreamStatus.STARTED

    @property
    def url(self) -> str | None:
        """Get the URL of the current stream.

        Returns:
            str | None: The URL of the current stream, or None if no stream is active.
        """
        return self._current.url if self._current else None

    async def start_stream(
        self,
        stream_type: StreamType,
        options: StreamOptions,
    ) -> ResultE[None]:
        """Start a stream of the specified type.

        Args:
            stream_type (StreamType): stream type to start
            options (StreamOptions): stream-specific options

        Returns:
            ResultE[None]: Result of the operation including failure reason if any.
        """
        if self.is_streaming:
            return ResultE.from_failure(GoProError(f"Stream {self.current_stream} is already active"))

        controller: StreamController
        match stream_type:
            case StreamType.WEBCAM:
                controller = WebcamStreamController(self._gopro)
            case StreamType.LIVE:
                controller = LiveStreamController(self._gopro)
            case StreamType.PREVIEW:
                controller = PreviewStreamController(self._gopro)

        if not controller.is_available:
            return ResultE.from_failure(GoProError(f"{stream_type} is not available"))

        match stream_type:
            case StreamType.WEBCAM:
                if not isinstance(options, WebcamStreamOptions):
                    return ResultE.from_failure(GoProError("Invalid options for webcam stream"))
            case StreamType.LIVE:
                if not isinstance(options, LivestreamOptions):
                    return ResultE.from_failure(GoProError("Invalid options for livestream"))
            case StreamType.PREVIEW:
                if not isinstance(options, PreviewStreamOptions):
                    return ResultE.from_failure(GoProError("Invalid options for preview stream"))

        self._current = controller
        logger.info(f"Starting {stream_type.name.title()} stream")
        match (result := await self._current.start(options)):
            case Success():
                logger.info("Stream started successfully.")
            case Failure():
                logger.error(f"Stream failed to start: {result.failure()}")
        return result

    async def stop_active_stream(self) -> ResultE[None]:
        """Stop the currently active stream.

        Returns:
            ResultE[None]: Result of the operation.
        """
        if self.is_streaming:
            assert self._current
            await self._current.stop()

        return ResultE.from_value(None)

    async def close(self) -> None:  # noqa: D102
        await self.stop_active_stream()



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
*
Commands related to controlling the camera's operation
"""

import builtins
import google.protobuf.descriptor
import google.protobuf.message
import sys

if sys.version_info >= (3, 8):
    import typing as typing_extensions
else:
    import typing_extensions
DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing_extensions.final
class RequestSetCameraName(google.protobuf.message.Message):
    """*
    Set the camera name

    Response: @ref ResponseGeneric
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    NAME_FIELD_NUMBER: builtins.int
    name: builtins.str
    "(TODO define encoding) string to identify the camera"

    def __init__(self, *, name: builtins.str | None = ...) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["name", b"name"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["name", b"name"]) -> None: ...

global___RequestSetCameraName = RequestSetCameraName



# util.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Thu Feb 23 22:43:52 UTC 2023

"""Common GUI utilities"""

from __future__ import annotations

import enum
import logging
import queue
import threading
from abc import ABC, abstractmethod
from typing import Any, Callable

import cv2

logger = logging.getLogger(__name__)


class StreamReader(ABC):
    """Stream Reader interface"""

    def __init__(self, source: str) -> None:
        self.source = source

    @abstractmethod
    def read(self) -> tuple[bool, Any]:
        """Read a frame from the stream

        Returns:
            tuple[bool, Any]: (success, frame)
        """

    @abstractmethod
    def display_frame(self, frame: Any) -> None:
        """Display a frame

        Args:
            frame (Any): frame to display
        """

    @abstractmethod
    def close(self) -> None:
        """Close the stream and release resources"""


class TsStreamReader(StreamReader):
    """Stream reader for TS streams

    Args:
        source (str): URL to read stream from
    """

    def __init__(self, source: str) -> None:
        super().__init__(source)
        self.cap = cv2.VideoCapture(source + "?overrun_nonfatal=1&fifo_size=50000000", cv2.CAP_FFMPEG)

    def read(self) -> tuple[bool, Any]:  # noqa
        return self.cap.read()

    def display_frame(self, frame: Any) -> None:  # noqa
        cv2.imshow(f"TS Stream from {self.source}", frame)

    def close(self) -> None:  # noqa
        self.cap.release()


class RtspStreamReader(StreamReader):
    """Stream reader for RTSP streams

    Args:
        source (str): URL to read stream from
    """

    def __init__(self, source: str) -> None:
        super().__init__(source)
        # TODO what arguments here?
        self.cap = cv2.VideoCapture(source)

    def read(self) -> tuple[bool, Any]:  # noqa
        return self.cap.read()

    def display_frame(self, frame: Any) -> None:  # noqa
        cv2.imshow(f"RTSP stream from {self.source}", frame)

    def close(self) -> None:  # noqa
        self.cap.release()


class BufferlessVideoCapture:
    """Buffer-less video capture to only display the most recent frame

    Open a video source to display it, and block until the user stops it by sending 'q'

    Args:
        source (str): video source to display
        protocol (Protocol): protocol to use to create the stream reader.
        printer (Callable): used to display output message. Defaults to print.
    """

    class Protocol(enum.Enum):
        """Stream Protocol"""

        RTSP = "RTSP"
        TS = "TS"

    def __init__(
        self,
        source: str,
        protocol: Protocol,
        printer: Callable = print,
    ) -> None:
        self._receive_thread = threading.Thread(
            target=self._read_frames,
            daemon=True,
            name="BufferlessVideoCaptureReceiveThread",
        )
        self.printer = printer
        self.printer(f"Starting viewer at {source}...")
        logger.debug(f"Starting viewer at {source}...")
        self._q: queue.Queue[Any] = queue.Queue()
        self._reader = TsStreamReader(source) if protocol == self.Protocol.TS else RtspStreamReader(source)
        # Start reading thread
        self._receive_thread.start()
        # TODO clean up thread on exit

    def display_blocking(self) -> None:
        """Display the stream, blocking, until the user stops it by sending 'q'"""
        self.printer("Viewer started")
        self.printer("Press 'q' in viewer to quit")
        while True:
            try:
                frame = self._q.get()
                self._reader.display_frame(frame)
                if cv2.waitKey(1) & 0xFF == ord("q"):
                    break
            except Exception as e:  # pylint: disable=broad-exception-caught
                logger.warning(e)
                break
        self._reader.close()
        cv2.destroyAllWindows()

    def _read_frames(self) -> None:
        """Read frames as soon as they are available, keeping only most recent one"""
        while True:
            try:
                ret, frame = self._reader.read()
                if not ret:
                    # logger.debug("Received empty frame.")
                    continue
                if not self._q.empty():
                    self._q.get_nowait()  # discard previous (unprocessed) frame
                self._q.put(frame)
            except queue.Empty:
                pass
            except Exception as e:  # pylint: disable=broad-exception-caught
                logger.error(f"Video capture received exception: {repr(e)}, stopping capture.")
                break



# webcam_stream.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Thu May 15 16:57:27 UTC 2025

"""Webcam stream controller implementation."""

from __future__ import annotations

import asyncio
import logging

from returns.result import ResultE

from open_gopro.domain.exceptions import GoProError
from open_gopro.domain.observable import Observable
from open_gopro.features.streaming.base_stream import StreamController, StreamType
from open_gopro.gopro_base import GoProBase
from open_gopro.models.constants import Toggle
from open_gopro.models.streaming import (
    WebcamError,
    WebcamProtocol,
    WebcamStatus,
    WebcamStreamOptions,
)

logger = logging.getLogger(__name__)


class WebcamStreamController(StreamController[WebcamStreamOptions]):
    """Livestream stream controller

    Args:
        gopro (GoProBase): GoPro device to operate on
    """

    def __init__(self, gopro: GoProBase) -> None:
        super().__init__(gopro)
        self.status_observable = Observable[StreamController.StreamStatus](
            debug_id="webcam stream controller status tracker"
        )
        self.current_options: WebcamStreamOptions | None = None
        self._status_task: asyncio.Task | None = None

    async def _track_status(self) -> None:
        """Track the webcam, converting it to a stream status observable"""
        # Poll until status is received
        # TODO cleanup?
        while True:
            response = (await self.gopro.http_command.webcam_status()).data
            if response.error != WebcamError.SUCCESS:
                # Something bad happened
                logger.error(f"Received webcam error: {response.error}")
                return  # TODO throw instead?
            logger.debug(f"Webcam stream controller received status: {response.status}")
            match response.status:
                case WebcamStatus.OFF:
                    await self.status_observable.emit(StreamController.StreamStatus.STOPPED)
                case WebcamStatus.IDLE:
                    await self.status_observable.emit(StreamController.StreamStatus.STOPPED)
                case WebcamStatus.HIGH_POWER_PREVIEW:
                    await self.status_observable.emit(StreamController.StreamStatus.STARTED)
                case WebcamStatus.LOW_POWER_PREVIEW:
                    await self.status_observable.emit(StreamController.StreamStatus.STARTED)
            await asyncio.sleep(1)

    @property
    def is_available(self) -> bool:  # noqa: D102
        return self.gopro.is_http_connected

    async def start(self, options: WebcamStreamOptions) -> ResultE[None]:  # noqa: D102
        if not self.is_available:
            logger.error("Webcam is not available")
            return ResultE.from_failure(GoProError("Webcam is not available"))
        if self.status in (StreamController.StreamStatus.STARTED, StreamController.StreamStatus.STARTING):
            logger.warning("Webcam is already started / starting")
            return ResultE.from_failure(GoProError("Webcam is already started / starting"))

        self._status_task = asyncio.create_task(self._track_status())
        self.current_options = options
        await self.status_observable.emit(StreamController.StreamStatus.STARTING)

        await self.gopro.http_command.set_shutter(shutter=Toggle.DISABLE)
        if (await self.gopro.http_command.webcam_status()).data.status not in {
            WebcamStatus.OFF,
            WebcamStatus.IDLE,
        }:
            logger.warning("Webcam is currently on. Turning if off.")
            assert (await self.gopro.http_command.webcam_stop()).ok
            await self.status_observable.observe().first(lambda s: s == StreamController.StreamStatus.STOPPED)

        logger.info("Starting webcam...")
        if (
            status := (await self.gopro.http_command.webcam_start(protocol=self.current_options.protocol)).data.error
        ) != WebcamError.SUCCESS:
            logger.error(f"Couldn't start webcam: {status}")
            self._cleanup()
            return ResultE.from_failure(GoProError(f"Couldn't start webcam: {status}"))
        await self.status_observable.observe().first(lambda s: s == StreamController.StreamStatus.STARTED)
        return ResultE.from_value(None)

    async def stop(self) -> ResultE[None]:  # noqa: D102
        if not self.is_available:
            logger.error("Webcam is not available")
            return ResultE.from_failure(GoProError("Webcam is not available"))
        if self.status == StreamController.StreamStatus.STOPPED:
            logger.warning("Webcam is already stopped")
            return ResultE.from_failure(GoProError("Webcam is already stopped"))

        await self.status_observable.emit(StreamController.StreamStatus.STOPPING)
        logger.info("Stopping webcam...")
        # First wait for it top stop
        assert (await self.gopro.http_command.webcam_stop()).ok
        await self.status_observable.observe().first(lambda s: s == StreamController.StreamStatus.STOPPED)
        # Now just tell it to exit
        await self.gopro.http_command.webcam_exit()
        self._cleanup()
        return ResultE.from_value(None)

    def _cleanup(self) -> None:
        """Cleanup after a stream has stopped or failed to start."""
        if self._status_task:
            self._status_task.cancel()
            self._status_task = None
        self.current_options = None

    @property
    def status(self) -> StreamController.StreamStatus:  # noqa: D102
        return self.status_observable.current or StreamController.StreamStatus.STOPPED

    @property
    def stream_type(self) -> StreamType:  # noqa: D102
        return StreamType.WEBCAM

    @property
    def url(self) -> str:  # noqa: D102
        assert self.current_options is not None, "Stream options not set"
        match self.current_options.protocol:
            case WebcamProtocol.RTSP:
                return f"rtsp://{self.gopro.ip_address}:554/live"
            case WebcamProtocol.TS | None:
                return r"udp://0.0.0.0:8554"

    async def close(self) -> ResultE[None]:  # noqa: D102
        if self._status_task:
            self._status_task.cancel()
        return ResultE.from_value(None)



# preview_stream.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Thu May 15 16:57:27 UTC 2025

"""Preview stream controller implementation."""

from __future__ import annotations

import logging

from returns.result import ResultE

from open_gopro.domain.exceptions import GoProError
from open_gopro.features.streaming.base_stream import StreamController, StreamType
from open_gopro.gopro_base import GoProBase
from open_gopro.models.constants import Toggle
from open_gopro.models.streaming import PreviewStreamOptions

logger = logging.getLogger(__name__)


# TODO is there a status we can track?
class PreviewStreamController(StreamController[PreviewStreamOptions]):
    """Preview stream controller

    Args:
        gopro (GoProBase): GoPro device to operate on
    """

    def __init__(self, gopro: GoProBase) -> None:
        super().__init__(gopro)
        self._status = StreamController.StreamStatus.STOPPED
        self._current_options: PreviewStreamOptions | None = None

    @property
    def is_available(self) -> bool:  # noqa: D102
        return self.gopro.is_http_connected

    async def start(self, options: PreviewStreamOptions) -> ResultE[None]:  # noqa: D102
        if not self.is_available:
            logger.error("Preview Stream is not available")
            return ResultE.from_failure(GoProError("Preview Stream is not available"))
        if self.status in (StreamController.StreamStatus.STARTED, StreamController.StreamStatus.STARTING):
            logger.warning("Preview Stream is already started / starting")
            return ResultE.from_failure(GoProError("Preview Stream is already started / starting"))

        self._current_options = options
        self._status = StreamController.StreamStatus.STARTING
        logger.info(f"Starting preview stream on port {self._current_options.port}")
        # Stop the preview stream if it is already running
        await self.gopro.http_command.set_preview_stream(mode=Toggle.DISABLE)
        await self.gopro.ble_command.set_shutter(shutter=Toggle.DISABLE)
        # Now start the preview stream
        response = await self.gopro.http_command.set_preview_stream(mode=Toggle.ENABLE, port=self._current_options.port)
        if response.ok:
            self._status = StreamController.StreamStatus.STARTED
            return ResultE.from_value(None)
        logger.error(f"Failed to start preview stream: {response.status}")
        self._status = StreamController.StreamStatus.STOPPED
        self._cleanup()
        return ResultE.from_failure(GoProError(f"Failed to start preview stream : {response.status}"))

    async def stop(self) -> ResultE[None]:  # noqa: D102
        if not self.is_available:
            logger.error("Preview Stream is not available")
            return ResultE.from_failure(GoProError("Preview Stream is not available"))
        if self.status == StreamController.StreamStatus.STOPPED:
            logger.warning("Preview Stream is already stopped")
            return ResultE.from_failure(GoProError("Preview Stream is already stopped"))

        self._status = StreamController.StreamStatus.STOPPING
        logger.info("Stopping preview stream")
        response = await self.gopro.http_command.set_preview_stream(mode=Toggle.DISABLE)
        self._cleanup()
        if response.ok:
            self._status = StreamController.StreamStatus.STOPPED
            return ResultE.from_value(None)
        logger.error(f"Failed to stop preview stream: {response.status}")
        return ResultE.from_failure(GoProError(f"Failed to stop preview stream : {response.status}"))

    def _cleanup(self) -> None:
        """Cleanup after a stream has stopped or failed to start."""
        self._current_options = None

    @property
    def status(self) -> StreamController.StreamStatus:  # noqa: D102
        return self._status

    @property
    def stream_type(self) -> StreamType:  # noqa: D102
        return StreamType.PREVIEW

    @property
    def url(self) -> str:  # noqa: D102
        assert self._current_options is not None, "Preview stream options not set"
        return f"udp://127.0.0.1:{self._current_options.port}"

    async def close(self) -> ResultE[None]:  # noqa: D102
        # Nothing to do
        return ResultE.from_value(None)



# __init__.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Thu May 15 16:57:27 UTC 2025

"""Top level package for the streaming feature."""

from .stream_feature import StreamFeature



# livestream.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Wed, Sep  1, 2021  5:05:45 PM

"""Example to start and view a livestream"""

import argparse
import asyncio

from rich.console import Console

from open_gopro import WirelessGoPro
from open_gopro.models import proto, streaming
from open_gopro.util import add_cli_args_and_parse, ainput
from open_gopro.util.logger import setup_logging

console = Console()  # rich consoler printer


async def main(args: argparse.Namespace) -> None:
    setup_logging(__name__, args.log)

    async with WirelessGoPro(
        args.identifier,
        enable_wifi=False,
        interfaces={WirelessGoPro.Interface.BLE},
    ) as gopro:
        console.print(f"[yellow]Connecting to {args.ssid}...")
        await gopro.access_point.connect(args.ssid, args.password)

        console.print(f"[yellow]Starting livestream to {args.url}...")
        await gopro.streaming.start_stream(
            streaming.StreamType.LIVE,
            streaming.LivestreamOptions(
                url=args.url,
                minimum_bitrate=args.min_bit,
                maximum_bitrate=args.max_bit,
                starting_bitrate=args.start_bit,
                resolution=args.resolution if args.resolution else None,
                fov=args.fov if args.fov else None,
                encode=args.encode,
            ),
        )
        assert gopro.streaming.url
        console.print(
            f"[yellow]Livestream to {gopro.streaming.url} is now streaming and should be available for viewing."
        )
        await ainput("Press enter to stop livestreaming...\n")

        console.print("[yellow]Stopping livestream...")
        await gopro.streaming.stop_active_stream()

        # TODO merge this into access point feature
        console.print("[yellow]Disconnecting GoPro from network...")
        await gopro.ble_command.release_network()

        console.print(" [yellow]Exiting...")


def parse_arguments() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description="Connect to the GoPro via BLE only, configure then start a Livestream, then display it with CV2."
    )
    parser.add_argument("ssid", type=str, help="WiFi SSID to connect to.")
    parser.add_argument("password", type=str, help="Password of WiFi SSID.")
    parser.add_argument("url", type=str, help="RTMP server URL to stream to.")
    parser.add_argument("--min_bit", type=int, help="Minimum bitrate.", default=1000)
    parser.add_argument("--max_bit", type=int, help="Maximum bitrate.", default=1000)
    parser.add_argument("--start_bit", type=int, help="Starting bitrate.", default=1000)
    parser.add_argument(
        "--resolution", help="Resolution.", choices=list(proto.EnumWindowSize.values()), default=None, type=int
    )
    parser.add_argument("--fov", help="Field of View.", choices=list(proto.EnumLens.values()), default=None, type=int)
    parser.add_argument("--encode", help="Save video to sdcard.", action=argparse.BooleanOptionalAction)
    parser.set_defaults(encode=True)
    return add_cli_args_and_parse(parser, wifi=False)


def entrypoint() -> None:
    asyncio.run(main(parse_arguments()))


if __name__ == "__main__":
    entrypoint()



# base_feature.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Mon May 12 23:03:50 UTC 2025

"""Functionality common to all features."""
from __future__ import annotations

import asyncio
from abc import ABC, abstractmethod
from typing import Any, Callable

import wrapt

from open_gopro.gopro_base import GoProBase


@wrapt.decorator
def require_supported(wrapped: Callable, instance: BaseFeature, args: Any, kwargs: Any) -> Any:
    """Ensure the feature is supported before allowing access.

    Args:
        wrapped (Callable): The wrapped function or property
        instance (BaseFeature): The class instance (or None if accessing a property getter)
        args (Any): Positional arguments
        kwargs (Any): Keyword arguments

    Returns:
        Any: Result of the wrapped function

    Raises:
        RuntimeError: If feature is not supported
    """
    # Handle the case where instance might be None (property access)
    # In this case, the instance is actually the first argument in args
    actual_instance = instance if instance is not None else args[0]

    if not actual_instance.is_supported:
        raise RuntimeError(f"{actual_instance.__class__.__name__} feature is not supported on this camera")

    if instance is None:
        # Property getter case - call it with the rest of args (skipping the first one which is the instance)
        if len(args) > 1:
            return wrapped(args[0], *args[1:], **kwargs)
        return wrapped(args[0])

    # Normal method call
    return wrapped(*args, **kwargs)


class BaseFeature(ABC):
    """Base Feature definition / interface"""

    def __init__(self) -> None:
        self._loop: asyncio.AbstractEventLoop
        self._gopro: GoProBase[Any]

    @property
    @abstractmethod
    def is_supported(self) -> bool:
        """Is the feature supported on the  current camera?

        Returns:
            bool: True if ready, False otherwise
        """

    async def open(  # pylint: disable=unused-argument
        self,
        loop: asyncio.AbstractEventLoop,
        gopro: GoProBase[Any],
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """Open the feature for use.

        This should be called before using any other methods in the feature. It must be called for each
        GoProBase instance that is used with the feature.

        Args:
            loop (asyncio.AbstractEventLoop): asyncio loop to use for this feature
            gopro (GoProBase[Any]): camera to operate on
            *args (Any): additional arguments for subclasses
            **kwargs (Any): additional keyword arguments for subclasses
        """
        self._loop = loop
        self._gopro = gopro

    @abstractmethod
    async def close(self) -> None:
        """Close the feature for use."""



# usb.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Fri Nov 18 00:18:13 UTC 2022

"""USB / wireless webcam demo"""

import argparse
import asyncio
import logging
import sys

from returns.pipeline import is_successful
from rich.console import Console

from open_gopro import WiredGoPro, WirelessGoPro
from open_gopro.demos.gui.video_display import BufferlessVideoCapture
from open_gopro.gopro_base import GoProBase
from open_gopro.models.streaming import StreamType, WebcamProtocol, WebcamStreamOptions
from open_gopro.util import add_cli_args_and_parse
from open_gopro.util.logger import set_stream_logging_level, setup_logging
from open_gopro.util.util import ainput

console = Console()


async def main(args: argparse.Namespace) -> int:
    logger = setup_logging(__name__, args.log)
    gopro: GoProBase | None = None

    try:
        wireless_interfaces: set[WirelessGoPro.Interface] = set()
        if args.wireless:
            wireless_interfaces = wireless_interfaces.union(
                {WirelessGoPro.Interface.BLE, WirelessGoPro.Interface.WIFI_AP}
            )
        async with (
            WirelessGoPro(
                args.identifier,
                host_wifi_interface=args.wifi_interface,
                interfaces=wireless_interfaces,
            )
            if wireless_interfaces
            else WiredGoPro(args.identifier)
        ) as gopro:
            set_stream_logging_level(logging.WARNING)
            with console.status(f"Starting webcam stream with protocol {args.protocol}..."):
                if not is_successful(
                    result := (
                        await gopro.streaming.start_stream(
                            stream_type=StreamType.WEBCAM,
                            options=WebcamStreamOptions(protocol=args.protocol),
                        )
                    )
                ):
                    console.print(f"[red]Failed to start webcam stream: {result.failure()}")
                    return 1

            url = gopro.streaming.url if args.protocol == WebcamProtocol.RTSP else "udp://@:8554"
            console.print(f"Preview stream started. It can be viewed in VLC at [yellow]{url}")
            console.print(
                "Press Enter to view the preview stream using Python CV2. Once started, it won't be viewable in VLC."
            )
            await ainput("")

            with console.status("Displaying the preview stream..."):
                assert gopro.streaming.url
                BufferlessVideoCapture(
                    source=gopro.streaming.url,
                    protocol=BufferlessVideoCapture.Protocol(args.protocol.name),
                    printer=console.print,
                ).display_blocking()

            with console.status("Stopping preview stream..."):
                await gopro.streaming.stop_active_stream()
        return 0

    except Exception as e:  # pylint: disable = broad-except
        logger.error(repr(e))

    if gopro:
        await gopro.close()

    return 0


def parse_arguments() -> argparse.Namespace:
    parser = argparse.ArgumentParser(description="Setup and view a GoPro webcam stream.")
    parser.add_argument(
        "--wireless",
        action="store_true",
        help="Set to use wireless (BLE / WIFI AP) instead of wired (USB) interface",
    )
    parser.add_argument(
        "--protocol",
        type=str,
        choices=["RTSP", "TS"],
        default="TS",
        help="Streaming protocol to use. Defaults to TS. RTSP is not supported on all GoPro models.",
    )
    args = add_cli_args_and_parse(parser)
    # Convert protocol string to enum if it is set
    match args.protocol.lower():
        case "rtsp":
            args.protocol = WebcamProtocol.RTSP
        case "ts":
            args.protocol = WebcamProtocol.TS
    return args


# Needed for poetry scripts defined in pyproject.toml
def entrypoint() -> None:
    sys.exit(asyncio.run(main(parse_arguments())))


if __name__ == "__main__":
    entrypoint()



# livestream.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Thu May 15 16:57:27 UTC 2025

"""Livestream stream controller implementation."""

from __future__ import annotations

import asyncio
import logging

from returns.result import ResultE

from open_gopro.domain.exceptions import GoProError
from open_gopro.domain.gopro_observable import GoProObservable
from open_gopro.domain.observable import Observable
from open_gopro.features.streaming.base_stream import StreamController, StreamType
from open_gopro.gopro_base import GoProBase
from open_gopro.models.constants import ActionId, Toggle
from open_gopro.models.constants.constants import FeatureId
from open_gopro.models.proto import (
    EnumLiveStreamStatus,
    EnumRegisterLiveStreamStatus,
    NotifyLiveStreamStatus,
)
from open_gopro.models.streaming import LivestreamOptions
from open_gopro.models.types import ProtobufId

logger = logging.getLogger(__name__)


class LiveStreamController(StreamController[LivestreamOptions]):
    """Livestream stream controller

    Args:
        gopro (GoProBase): GoPro device to operate on
    """

    def __init__(self, gopro: GoProBase) -> None:
        super().__init__(gopro)
        self.status_observable = Observable[StreamController.StreamStatus](
            debug_id="livestream controller status tracker"
        )
        self.current_options: LivestreamOptions | None = None
        self._status_task = asyncio.create_task(self._track_status())

    async def _track_status(self) -> None:
        """Track the status of the livestream status observable, converting it to a stream status observable."""
        async for status in (await self.get_livestream_status_observable()).observe(
            debug_id="Livestream Status Tracker"
        ):
            logger.debug(f"Livestream controller received status: {status}")
            match status.live_stream_status:
                case (
                    EnumLiveStreamStatus.LIVE_STREAM_STATE_IDLE
                    | EnumLiveStreamStatus.LIVE_STREAM_STATE_FAILED_STAY_ON
                    | EnumLiveStreamStatus.LIVE_STREAM_STATE_UNAVAILABLE
                ):
                    logger.debug(f"Livestream controller emitting state {StreamController.StreamStatus.STOPPED}")
                    await self.status_observable.emit(StreamController.StreamStatus.STOPPED)
                case EnumLiveStreamStatus.LIVE_STREAM_STATE_READY | EnumLiveStreamStatus.LIVE_STREAM_STATE_RECONNECTING:
                    logger.debug(f"Livestream controller emitting state {StreamController.StreamStatus.STARTING}")
                    await self.status_observable.emit(StreamController.StreamStatus.STARTING)
                case (
                    EnumLiveStreamStatus.LIVE_STREAM_STATE_STREAMING
                    | EnumLiveStreamStatus.LIVE_STREAM_STATE_COMPLETE_STAY_ON
                ):
                    logger.debug(f"Livestream controller emitting state {StreamController.StreamStatus.STARTED}")
                    await self.status_observable.emit(StreamController.StreamStatus.STARTED)
                case _:
                    logger.warning(f"Livestream controller received unknown status: {status}")

    async def get_livestream_status_observable(
        self,
    ) -> GoProObservable[NotifyLiveStreamStatus]:
        """Get an API-level observable for the livestream status protobuf operation.

        Returns:
            GoProObservable[NotifyLiveStreamStatus]: live stream status observable
        """
        return await GoProObservable[NotifyLiveStreamStatus](
            gopro=self.gopro,
            register_command=self.gopro.ble_command.register_livestream_status(
                register=[EnumRegisterLiveStreamStatus.REGISTER_LIVE_STREAM_STATUS_STATUS]
            ),
            unregister_command=self.gopro.ble_command.register_livestream_status(
                unregister=[EnumRegisterLiveStreamStatus.REGISTER_LIVE_STREAM_STATUS_STATUS]
            ),
            update=ProtobufId(FeatureId.QUERY, ActionId.LIVESTREAM_STATUS_NOTIF),
        ).start()

    @property
    def is_available(self) -> bool:  # noqa: D102
        # TODO can we check if the camera is connected to an access point? We probably need to update the AP feature.
        return self.gopro.is_ble_connected

    async def start(self, options: LivestreamOptions) -> ResultE[None]:  # noqa: D102
        if not self.is_available:
            logger.error("Livestream is not available")
            return ResultE.from_failure(GoProError("Livestream is not available"))
        if self.status in (StreamController.StreamStatus.STARTED, StreamController.StreamStatus.STARTING):
            logger.warning("Livestream is already started / starting")
            return ResultE.from_failure(GoProError("Livestream is already started / starting"))

        logger.info("Starting livestream")
        self.current_options = options
        # Disable shutter before starting livestream
        # TODO error handling
        await self.gopro.ble_command.set_shutter(shutter=Toggle.DISABLE)
        assert (
            await self.gopro.ble_command.set_livestream_mode(
                url=self.current_options.url,
                window_size=self.current_options.resolution,
                minimum_bitrate=self.current_options.minimum_bitrate,
                maximum_bitrate=self.current_options.maximum_bitrate,
                starting_bitrate=self.current_options.starting_bitrate,
                encode=self.current_options.encode,
                lens=self.current_options.fov,
            )
        ).ok
        logger.info("Waiting for livestream to begin starting...")
        await self.status_observable.observe(debug_id="Livestream Controller Wait For Ready").first(
            lambda s: s == StreamController.StreamStatus.STARTING
        )
        await asyncio.sleep(2)  # TODO Is this still needed?
        logger.info("Setting shutter to start livestream")
        await self.gopro.ble_command.set_shutter(shutter=Toggle.ENABLE)
        return ResultE.from_value(None)

    async def stop(self) -> ResultE[None]:  # noqa: D102
        if not self.is_available:
            logger.error("Livestream is not available")
            return ResultE.from_failure(GoProError("Livestream is not available"))
        if self.status == StreamController.StreamStatus.STOPPED:
            logger.warning("Livestream is already stopped")
            return ResultE.from_failure(GoProError("Livestream is already stopped"))

        logger.info("Stopping livestream")
        # TODO error handling
        await self.gopro.ble_command.set_shutter(shutter=Toggle.DISABLE)
        return ResultE.from_value(None)

    @property
    def status(self) -> StreamController.StreamStatus:  # noqa: D102
        return self.status_observable.current or StreamController.StreamStatus.STOPPED

    @property
    def stream_type(self) -> StreamType:  # noqa: D102
        return StreamType.LIVE

    @property
    def url(self) -> str:  # noqa: D102
        if self.current_options is None:
            raise GoProError("Livestream is not started")
        return self.current_options.url

    def _cleanup(self) -> None:
        """Cleanup after a stream has stopped or failed to start."""
        self.current_options = None

    async def close(self) -> ResultE[None]:  # noqa: D102
        self._status_task.cancel()
        return ResultE.from_value(None)
